"""
Helper for tests
"""

import asyncio

def async_test(func):

    def wrapper(*args, **kwargs):
        coro = asyncio.coroutine(func)
        future = coro(*args, **kwargs)
        loop = asyncio.get_event_loop()
        loop.run_until_complete(future)
    return wrapper



import asyncio
import sys

from pyramid.interfaces import (
    IExceptionViewClassifier,
    IRequest,
    IView,
    )

from zope.interface import providedBy

from pyramid.config import tweens

from .aioinspect import is_generator


class Tweens(tweens.Tweens):

    def __call__(self, handler, registry):

        def handle_handler(handler):
            @asyncio.coroutine
            def handle_request(request):
                response = handler(request)
                if is_generator(response):
                    response = yield from response
                return response
            return handle_request

        if self.explicit:
            use = self.explicit
        else:
            use = self.implicit()

        for name, factory in use[::-1]:
            handler = factory(handler, registry)
        return handler




import asyncio
import logging
from collections import defaultdict

from zope.interface import implementer
from pyramid.interfaces import ISession, ISessionFactory

from .kvs import session, serializer


log = logging.getLogger(__name__)

@implementer(ISession)
class CookieSession(session.CookieSession):

    def __init__(self, request, client, key_name):
        self._dirty = False
        self.key_name = key_name
        self.client = client
        self.request = request

        self._session_key = self.get_session_key()
        self._session_data = None

    @asyncio.coroutine
    def load_session_data(self):
        self._session_data = defaultdict(defaultdict)

        if not self._session_key:
            log.warn('No session found')
            return

        stored_data = yield from self.client.get(self._session_key)
        if stored_data:
            self._session_data.update(stored_data)
        else:
            self.changed()

    def get_session_key(self):
        session_key = self.request.cookies.get(self.key_name)
        if not session_key:
            session_key = session._create_token()  # XXX private method called
        return session_key

    @asyncio.coroutine
    def save_session(self, request, response):
        if self._session_data is None:  # session invalidated
            self.client.delete(self._session_key)
            response.delete_cookie(self.key_name)
            return
        response.set_cookie(self.key_name, self._session_key,
                            self.client.ttl)
        yield from self.client.set(self._session_key, self._session_data)


@implementer(ISessionFactory)
class SessionFactory(object):
    session_class = CookieSession

    def __init__(self, settings):
        self.config = serializer('json').loads(settings['asyncio.session'])
        self.config.setdefault('key_prefix', 'session::')
        self.key_name = self.config.pop('key_name', 'session_id')
        self._client = None

    @asyncio.coroutine
    def __call__(self, request):
        if self._client is None:
            self.config['kvs'] = 'aioredis'
            self._client = kvs.KVS(**self.config)
            self._client._client = yield from self._client._client
        session = CookieSession(request, self._client, self.key_name)
        yield from session.load_session_data()
        return session



"""
Override the Pyramid Router in order to get request handler as
asyncio coroutine.
"""

import sys
import asyncio
import logging

from zope.interface import (
    implementer,
    providedBy,
    )

from pyramid.interfaces import (
    IRequest,
    IRouteRequest,
    IRouter,
    ISessionFactory,
    ITraverser,
    IView,
    IViewClassifier,
    IExceptionViewClassifier,
    ITweens,
    )

from pyramid.events import (
    ContextFound,
    NewRequest,
    NewResponse,
    )

from pyramid.exceptions import PredicateMismatch, ConfigurationError
from pyramid.httpexceptions import HTTPException, HTTPNotFound
from pyramid.settings import aslist

from pyramid.traversal import (
    ResourceTreeTraverser,
    )

from pyramid.router import Router as RouterBase
from .tweens import Tweens
from pyramid_asyncio.aioinspect import is_generator

log = logging.getLogger(__name__)


@implementer(IRouter)
class Router(RouterBase):

    def __init__(self, config):
        self.first_route = True
        self.config = config
        super().__init__(config.registry)

    @asyncio.coroutine
    def handle_request(self, request):

        attrs = request.__dict__
        registry = attrs['registry']

        request.request_iface = IRequest
        context = None
        routes_mapper = self.routes_mapper
        debug_routematch = self.debug_routematch
        adapters = registry.adapters
        has_listeners = registry.has_listeners
        notify = registry.notify
        logger = self.logger
        if request.registry.queryUtility(ISessionFactory) is not None and is_generator(request.session):
            request.session = yield from request.session

        has_listeners and notify(NewRequest(request))
        # find the root object
        root_factory = self.root_factory
        if routes_mapper is not None:
            info = routes_mapper(request)
            match, route = info['match'], info['route']
            if route is None:
                if debug_routematch:
                    msg = ('no route matched for url %s' %
                           request.url)
                    logger and logger.debug(msg)
            else:
                attrs['matchdict'] = match
                attrs['matched_route'] = route

                if debug_routematch:
                    msg = (
                        'route matched for url %s; '
                        'route_name: %r, '
                        'path_info: %r, '
                        'pattern: %r, '
                        'matchdict: %r, '
                        'predicates: %r' % (
                            request.url,
                            route.name,
                            request.path_info,
                            route.pattern,
                            match,
                            ', '.join([p.text() for p in route.predicates]))
                        )
                    logger and logger.debug(msg)

                request.request_iface = registry.queryUtility(
                    IRouteRequest,
                    name=route.name,
                    default=IRequest)

                root_factory = route.factory or self.root_factory

        root = root_factory(request)
        attrs['root'] = root

        # find a context
        traverser = adapters.queryAdapter(root, ITraverser)
        if traverser is None:
            traverser = ResourceTreeTraverser(root)
        tdict = traverser(request)

        context, view_name, subpath, traversed, vroot, vroot_path = (
            tdict['context'],
            tdict['view_name'],
            tdict['subpath'],
            tdict['traversed'],
            tdict['virtual_root'],
            tdict['virtual_root_path']
            )

        attrs.update(tdict)
        has_listeners and notify(ContextFound(request))

        # find a view callable
        context_iface = providedBy(context)
        view_callable = adapters.lookup(
            (IViewClassifier, request.request_iface, context_iface),
            IView, name=view_name, default=None)

        # invoke the view callable
        if view_callable is None:
            if self.debug_notfound:
                msg = (
                    'debug_notfound of url %s; path_info: %r, '
                    'context: %r, view_name: %r, subpath: %r, '
                    'traversed: %r, root: %r, vroot: %r, '
                    'vroot_path: %r' % (
                        request.url, request.path_info, context,
                        view_name, subpath, traversed, root, vroot,
                        vroot_path)
                    )
                logger and logger.debug(msg)
            else:
                msg = request.path_info
            raise HTTPNotFound(msg)
        else:
            try:
                if asyncio.iscoroutinefunction(view_callable):
                    response = yield from view_callable(context, request)
                else:
                    response = view_callable(context, request)
                while is_generator(response):
                    response = yield from response
            except PredicateMismatch:
                # look for other views that meet the predicate
                # criteria
                for iface in context_iface.__sro__[1:]:
                    previous_view_callable = view_callable
                    view_callable = adapters.lookup(
                        (IViewClassifier, request.request_iface, iface),
                        IView, name=view_name, default=None)
                    # intermediate bases may lookup same view_callable
                    if view_callable is previous_view_callable:
                        continue
                    if view_callable is not None:
                        try:
                            response = yield from view_callable(context,
                                                                request)
                            break
                        except PredicateMismatch:
                            pass
                else:
                    raise
        return response

    @asyncio.coroutine
    def invoke_subrequest(self, request, use_tweens=False):
        """Obtain a response object from the Pyramid application based on
        information in the ``request`` object provided.  The ``request``
        object must be an object that implements the Pyramid request
        interface (such as a :class:`pyramid.request.Request` instance).  If
        ``use_tweens`` is ``True``, the request will be sent to the
        :term:`tween` in the tween stack closest to the request ingress.  If
        ``use_tweens`` is ``False``, the request will be sent to the main
        router handler, and no tweens will be invoked.

        See the API for pyramid.request for complete documentation.
        """
        registry = self.registry
        has_listeners = self.registry.has_listeners
        notify = self.registry.notify
        threadlocals = {'registry': registry, 'request': request}
        manager = self.threadlocal_manager
        manager.push(threadlocals)
        request.registry = registry
        request.invoke_subrequest = self.invoke_subrequest

        if use_tweens:
            # XXX Recopy tweens state, registered my own ITweens does not
            # save the registred handler. Should invest more
            tween = Tweens()
            registred_tweens = registry.queryUtility(ITweens)
            if registred_tweens is not None:
                tween.explicit = registred_tweens.explicit
                tween.implicit = registred_tweens.implicit
                handle_request = tween(self.orig_handle_request, registry)
            else:
                handle_request = self.orig_handle_request
        else:
            handle_request = self.orig_handle_request

        try:

            try:
                extensions = self.request_extensions
                if extensions is not None:
                    request._set_extensions(extensions)

                response = yield from handle_request(request)

                if request.response_callbacks:
                    yield from request._process_response_callbacks(response)

                has_listeners and notify(NewResponse(request, response))

            # XXX excview_tween_factory is not a generator
            # Move a part of its code here
            except Exception as exc:
                # WARNING: do not assign the result of sys.exc_info() to a local
                # var here, doing so will cause a leak.  We used to actually
                # explicitly delete both "exception" and "exc_info" from ``attrs``
                # in a ``finally:`` clause below, but now we do not because these
                # attributes are useful to upstream tweens.  This actually still
                # apparently causes a reference cycle, but it is broken
                # successfully by the garbage collector (see
                # https://github.com/Pylons/pyramid/issues/1223).
                attrs= request.__dict__
                attrs['exc_info'] = sys.exc_info()
                attrs['exception'] = exc
                adapters = request.registry.adapters

                # clear old generated request.response, if any; it may
                # have been mutated by the view, and its state is not
                # sane (e.g. caching headers)
                if 'response' in attrs:
                    del attrs['response']
                # we use .get instead of .__getitem__ below due to
                # https://github.com/Pylons/pyramid/issues/700
                request_iface = attrs.get('request_iface', IRequest)
                provides = providedBy(exc)
                for_ = (IExceptionViewClassifier, request_iface.combined, provides)
                view_callable = adapters.lookup(for_, IView, default=None)
                if view_callable is None:
                    raise
                response = view_callable(exc, request)
                while is_generator(response):
                    response = yield from response

            finally:
                if request.finished_callbacks:
                    yield from request._process_finished_callbacks()

        finally:
            manager.pop()

        return response


    @asyncio.coroutine
    def __call__(self, environ, start_response):
        """
        Accept ``environ`` and ``start_response``; create a
        :term:`request` and route the request to a :app:`Pyramid`
        view based on introspection of :term:`view configuration`
        within the application registry; call ``start_response`` and
        return an iterable.
        """
        request = yield from self.request_factory(environ)
        try:
            response = yield from self.invoke_subrequest(request,
                                                         use_tweens=True)
        except HTTPException as exc:
            response = exc
        response = response(request.environ, start_response)
        return response

    exit_handlers = []

    @asyncio.coroutine
    def open(self):
        settings = self.config.get_settings()
        aioincludes = aslist(settings.get('asyncio.includes', ''))

        for callable in aioincludes:
            try:
                module = self.config.maybe_dotted(callable)
                try:
                    includeme = getattr(module, 'includeme')
                except AttributeError:
                    raise ConfigurationError(
                        "module %r has no attribute 'includeme'" % (module.__name__)
                        )

                yield from includeme(self.config)
            except Exception:
                log.exception('{} raise an exception'.format(callable))
        self.config.commit()

    @asyncio.coroutine
    def close(self):
        for handler in self.exit_handlers:
            yield from handler()


def add_exit_handler(config, handler):
    Router.exit_handlers.append(handler)



import venusian


class coroutine_view_config(object):
    """
    A patched version of view_config that use coroutine for views
    """
    venusian = venusian # for testing injection
    def __init__(self, **settings):
        if 'for_' in settings:
            if settings.get('context') is None:
                settings['context'] = settings['for_']
        self.__dict__.update(settings)

    def __call__(self, wrapped):
        settings = self.__dict__.copy()
        depth = settings.pop('_depth', 0)

        def callback(context, name, ob):
            config = context.config.with_package(info.module)
            config.add_coroutine_view(view=ob, **settings)

        info = self.venusian.attach(wrapped, callback, category='pyramid',
                                    depth=depth + 1)

        if info.scope == 'class':
            # if the decorator was attached to a method in a class, or
            # otherwise executed at class scope, we need to set an
            # 'attr' into the settings if one isn't already in there
            if settings.get('attr') is None:
                settings['attr'] = wrapped.__name__

        settings['_info'] = info.codeinfo # fbo "action_method"
        return wrapped



import asyncio
from io import BytesIO

from zope.interface import implementer
from pyramid.interfaces import (
    IRequestFactory,
    IRequest,
    IAuthenticationPolicy,
    IAuthorizationPolicy,
    )

from pyramid.request import Request as BaseRequest
from pyramid.security import Allowed
from .aioinspect import is_generator


@implementer(IRequest)
class Request(BaseRequest):

    @asyncio.coroutine
    def _process_response_callbacks(self, response):
        callbacks = self.response_callbacks
        while callbacks:
            callback = callbacks.pop(0)
            callback_resp = callback(self, response)
            if is_generator(callback_resp):
                yield from callback_resp

    @asyncio.coroutine
    def _process_finished_callbacks(self):
        callbacks = self.finished_callbacks
        while callbacks:
            callback = callbacks.pop(0)
            callback_resp = callback(self)
            if is_generator(callback_resp):
                yield from callback_resp

    @asyncio.coroutine
    def has_permission(self, permission, context=None):
        if context is None:
            context = self.context
        reg = self.registry
        authn_policy = reg.queryUtility(IAuthenticationPolicy)
        if authn_policy is None:
            return Allowed('No authentication policy in use.')
        authz_policy = reg.queryUtility(IAuthorizationPolicy)
        if authz_policy is None:
            raise ValueError('Authentication policy registered without '
                             'authorization policy') # should never happen
        principals = authn_policy.effective_principals(self)
        if is_generator(principals):
            principals = yield from principals
        permits = authz_policy.permits(context, principals, permission)
        if is_generator(permits):
            permits = yield from permits
        return permits


@implementer(IRequestFactory)
class RequestFactory:
    """ A utility which generates a request """

    @asyncio.coroutine
    def __call__(self, environ):
        """ Return an object implementing IRequest, e.g. an instance
        of ``pyramid.request.Request``"""

        body = environ['wsgi.input'].read()
        if is_generator(body):
            # wsgi.input can be a generator depending of
            # the configuration of aiohttp
            body = yield from body 
        environ['wsgi.input'] = BytesIO(body)
        return Request(environ)

    @classmethod
    def blank(cls, path):
        """ Return an empty request object (see
        :meth:`pyramid.request.Request.blank`)"""
        return Request.blank(path)



"""
This code is a big copy/paste of code from pyramid and change the
view in order to handle it as a coroutine
"""
import warnings
import asyncio
import inspect

from zope.interface import Interface, implementedBy, implementer
from zope.interface.interfaces import IInterface
from pyramid import renderers
from pyramid.config import Configurator as ConfiguratorBase, global_registries
from pyramid.config.views import (ViewDeriver as ViewDeriverBase,
                                  StaticURLInfo as StaticURLInfoBase,
                                  isexception,
                                  wraps_view,
                                  view_description)
from pyramid.config.util import DEFAULT_PHASH, MAX_ORDER
from pyramid.events import ApplicationCreated
from pyramid.util import action_method, viewdefaults
from pyramid.config.views import MultiView
from pyramid.security import NO_PERMISSION_REQUIRED
from pyramid.response import Response
from pyramid.compat import string_types, is_nonstr_iter
from pyramid.registry import predvalseq, Deferred
from pyramid.exceptions import ConfigurationError
from pyramid.interfaces import (IDefaultPermission, IRequest, IRouteRequest,
                                IView, ISecuredView, IMultiView,
                                IResponse,
                                IRendererFactory, IViewClassifier,
                                IExceptionResponse, IExceptionViewClassifier,
                                )
from pyramid.httpexceptions import HTTPForbidden
from .router import Router, add_exit_handler
from .request import RequestFactory
from .aioinspect import is_generator


@viewdefaults
@action_method
def add_coroutine_view(
    config,
    view=None,
    name="",
    for_=None,
    permission=None,
    request_type=None,
    route_name=None,
    request_method=None,
    request_param=None,
    containment=None,
    attr=None,
    renderer=None,
    wrapper=None,
    xhr=None,
    accept=None,
    header=None,
    path_info=None,
    custom_predicates=(),
    context=None,
    decorator=None,
    mapper=None,
    http_cache=None,
    match_param=None,
    check_csrf=None,
    **predicates):
    """ patched version of pyramid add_view that use asyncio coroutine """
    self = config
    if custom_predicates:
        warnings.warn(
            ('The "custom_predicates" argument to Configurator.add_view '
             'is deprecated as of Pyramid 1.5.  Use '
             '"config.add_view_predicate" and use the registered '
             'view predicate as a predicate argument to add_view instead. '
             'See "Adding A Third Party View, Route, or Subscriber '
             'Predicate" in the "Hooks" chapter of the documentation '
             'for more information.'),
            DeprecationWarning,
            stacklevel=4
            )

    view = self.maybe_dotted(view)
    # transform the view to a coroutine only in case it's really a coroutine
    if not asyncio.iscoroutinefunction(view) and is_generator(view):
        view = asyncio.coroutine(view)

    context = self.maybe_dotted(context)
    for_ = self.maybe_dotted(for_)
    containment = self.maybe_dotted(containment)
    mapper = self.maybe_dotted(mapper)

    def combine(*decorators):
        def decorated(view_callable):
            # reversed() is allows a more natural ordering in the api
            for decorator in reversed(decorators):
                view_callable = decorator(view_callable)
            return view_callable
        return decorated

    if is_nonstr_iter(decorator):
        decorator = combine(*map(self.maybe_dotted, decorator))
    else:
        decorator = self.maybe_dotted(decorator)

    if not view:
        if renderer:
            def view(context, request):
                return {}
        else:
            raise ConfigurationError('"view" was not specified and '
                                     'no "renderer" specified')

    if request_type is not None:
        request_type = self.maybe_dotted(request_type)
        if not IInterface.providedBy(request_type):
            raise ConfigurationError(
                'request_type must be an interface, not %s' % request_type)

    if context is None:
        context = for_

    r_context = context
    if r_context is None:
        r_context = Interface
    if not IInterface.providedBy(r_context):
        r_context = implementedBy(r_context)

    if isinstance(renderer, string_types):
        renderer = renderers.RendererHelper(
            name=renderer, package=self.package,
            registry = self.registry)

    if accept is not None:
        accept = accept.lower()

    introspectables = []
    pvals = predicates.copy()
    pvals.update(
        dict(
            xhr=xhr,
            request_method=request_method,
            path_info=path_info,
            request_param=request_param,
            header=header,
            accept=accept,
            containment=containment,
            request_type=request_type,
            match_param=match_param,
            check_csrf=check_csrf,
            custom=predvalseq(custom_predicates),
            )
        )

    def discrim_func():
        # We need to defer the discriminator until we know what the phash
        # is.  It can't be computed any sooner because thirdparty
        # predicates may not yet exist when add_view is called.
        order, preds, phash = predlist.make(self, **pvals)
        view_intr.update({'phash':phash, 'order':order, 'predicates':preds})
        return ('view', context, name, route_name, phash)

    discriminator = Deferred(discrim_func)

    if inspect.isclass(view) and attr:
        view_desc = 'method %r of %s' % (
            attr, self.object_description(view))
    else:
        view_desc = self.object_description(view)

    tmpl_intr = None

    view_intr = self.introspectable('views',
                                    discriminator,
                                    view_desc,
                                    'view')
    view_intr.update(
        dict(name=name,
             context=context,
             containment=containment,
             request_param=request_param,
             request_methods=request_method,
             route_name=route_name,
             attr=attr,
             xhr=xhr,
             accept=accept,
             header=header,
             path_info=path_info,
             match_param=match_param,
             check_csrf=check_csrf,
             callable=view,
             mapper=mapper,
             decorator=decorator,
             )
        )
    view_intr.update(**predicates)
    introspectables.append(view_intr)
    predlist = self.get_predlist('view')

    def register(permission=permission, renderer=renderer):
        # the discrim_func above is guaranteed to have been called already
        order = view_intr['order']
        preds = view_intr['predicates']
        phash = view_intr['phash']
        request_iface = IRequest
        if route_name is not None:
            request_iface = self.registry.queryUtility(IRouteRequest,
                                                       name=route_name)
            if request_iface is None:
                # route configuration should have already happened in
                # phase 2
                raise ConfigurationError(
                    'No route named %s found for view registration' %
                    route_name)

        if renderer is None:
            # use default renderer if one exists (reg'd in phase 1)
            if self.registry.queryUtility(IRendererFactory) is not None:
                renderer = renderers.RendererHelper(
                    name=None,
                    package=self.package,
                    registry=self.registry
                    )

        if permission is None:
            # intent: will be None if no default permission is registered
            # (reg'd in phase 1)
            permission = self.registry.queryUtility(IDefaultPermission)

        # added by discrim_func above during conflict resolving
        preds = view_intr['predicates']
        order = view_intr['order']
        phash = view_intr['phash']

        # __no_permission_required__ handled by _secure_view
        deriver = ViewDeriver(
            registry=self.registry,
            permission=permission,
            predicates=preds,
            attr=attr,
            renderer=renderer,
            wrapper_viewname=wrapper,
            viewname=name,
            accept=accept,
            order=order,
            phash=phash,
            package=self.package,
            mapper=mapper,
            decorator=decorator,
            http_cache=http_cache,
            )
        derived_view = deriver(view)
        derived_view.__discriminator__ = lambda *arg: discriminator
        # __discriminator__ is used by superdynamic systems
        # that require it for introspection after manual view lookup;
        # see also MultiView.__discriminator__
        view_intr['derived_callable'] = derived_view

        registered = self.registry.adapters.registered

        # A multiviews is a set of views which are registered for
        # exactly the same context type/request type/name triad.  Each
        # consituent view in a multiview differs only by the
        # predicates which it possesses.

        # To find a previously registered view for a context
        # type/request type/name triad, we need to use the
        # ``registered`` method of the adapter registry rather than
        # ``lookup``.  ``registered`` ignores interface inheritance
        # for the required and provided arguments, returning only a
        # view registered previously with the *exact* triad we pass
        # in.

        # We need to do this three times, because we use three
        # different interfaces as the ``provided`` interface while
        # doing registrations, and ``registered`` performs exact
        # matches on all the arguments it receives.

        old_view = None

        for view_type in (IView, ISecuredView, IMultiView):
            old_view = registered((IViewClassifier, request_iface,
                                   r_context), view_type, name)
            if old_view is not None:
                break

        isexc = isexception(context)

        def regclosure():
            if hasattr(derived_view, '__call_permissive__'):
                view_iface = ISecuredView
            else:
                view_iface = IView
            self.registry.registerAdapter(
                derived_view,
                (IViewClassifier, request_iface, context), view_iface, name
                )
            if isexc:
                self.registry.registerAdapter(
                    derived_view,
                    (IExceptionViewClassifier, request_iface, context),
                    view_iface, name)

        is_multiview = IMultiView.providedBy(old_view)
        old_phash = getattr(old_view, '__phash__', DEFAULT_PHASH)

        if old_view is None:
            # - No component was yet registered for any of our I*View
            #   interfaces exactly; this is the first view for this
            #   triad.
            regclosure()

        elif (not is_multiview) and (old_phash == phash):
            # - A single view component was previously registered with
            #   the same predicate hash as this view; this registration
            #   is therefore an override.
            regclosure()

        else:
            # - A view or multiview was already registered for this
            #   triad, and the new view is not an override.

            # XXX we could try to be more efficient here and register
            # a non-secured view for a multiview if none of the
            # multiview's consituent views have a permission
            # associated with them, but this code is getting pretty
            # rough already
            if is_multiview:
                multiview = old_view
            else:
                multiview = MultiView(name)
                old_accept = getattr(old_view, '__accept__', None)
                old_order = getattr(old_view, '__order__', MAX_ORDER)
                multiview.add(old_view, old_order, old_accept, old_phash)
            multiview.add(derived_view, order, accept, phash)
            for view_type in (IView, ISecuredView):
                # unregister any existing views
                self.registry.adapters.unregister(
                    (IViewClassifier, request_iface, r_context),
                    view_type, name=name)
                if isexc:
                    self.registry.adapters.unregister(
                        (IExceptionViewClassifier, request_iface,
                         r_context), view_type, name=name)
            self.registry.registerAdapter(
                multiview,
                (IViewClassifier, request_iface, context),
                IMultiView, name=name)
            if isexc:
                self.registry.registerAdapter(
                    multiview,
                    (IExceptionViewClassifier, request_iface, context),
                    IMultiView, name=name)
        renderer_type = getattr(renderer, 'type', None) # gard against None
        intrspc = self.introspector
        if (
            renderer_type is not None and
            tmpl_intr is not None and
            intrspc is not None and
            intrspc.get('renderer factories', renderer_type) is not None
            ):
            # allow failure of registered template factories to be deferred
            # until view execution, like other bad renderer factories; if
            # we tried to relate this to an existing renderer factory
            # without checking if it the factory actually existed, we'd end
            # up with a KeyError at startup time, which is inconsistent
            # with how other bad renderer registrations behave (they throw
            # a ValueError at view execution time)
            tmpl_intr.relate('renderer factories', renderer.type)

    if mapper:
        mapper_intr = self.introspectable(
            'view mappers',
            discriminator,
            'view mapper for %s' % view_desc,
            'view mapper'
            )
        mapper_intr['mapper'] = mapper
        mapper_intr.relate('views', discriminator)
        introspectables.append(mapper_intr)
    if route_name:
        view_intr.relate('routes', route_name) # see add_route
    if renderer is not None and renderer.name and '.' in renderer.name:
        # the renderer is a template
        tmpl_intr = self.introspectable(
            'templates',
            discriminator,
            renderer.name,
            'template'
            )
        tmpl_intr.relate('views', discriminator)
        tmpl_intr['name'] = renderer.name
        tmpl_intr['type'] = renderer.type
        tmpl_intr['renderer'] = renderer
        introspectables.append(tmpl_intr)
    if permission is not None:
        # if a permission exists, register a permission introspectable
        perm_intr = self.introspectable(
            'permissions',
            permission,
            permission,
            'permission'
            )
        perm_intr['value'] = permission
        perm_intr.relate('views', discriminator)
        introspectables.append(perm_intr)
    self.action(discriminator, register, introspectables=introspectables)


def make_asyncio_app(config):
    self = config
    self.set_request_factory(RequestFactory())
    self.commit()
    app = Router(self)

    # Allow tools like "pshell development.ini" to find the 'last'
    # registry configured.
    global_registries.add(self.registry)

    # Push the registry onto the stack in case any code that depends on
    # the registry threadlocal APIs used in listeners subscribed to the
    # IApplicationCreated event.
    self.manager.push({'registry':self.registry, 'request':None})
    try:
        self.registry.notify(ApplicationCreated(app))
    finally:
        self.manager.pop()

    return app



class ViewDeriver(ViewDeriverBase):

    @wraps_view
    def secured_view(self, view):
        permission = self.kw.get('permission')
        if permission == NO_PERMISSION_REQUIRED:
            # allow views registered within configurations that have a
            # default permission to explicitly override the default
            # permission, replacing it with no permission at all
            permission = None

        wrapped_view = view
        if self.authn_policy and self.authz_policy and (permission is not None):
            @asyncio.coroutine
            def _permitted(context, request):
                principals = self.authn_policy.effective_principals(request)
                if is_generator(principals):
                    principals = yield from principals
                return self.authz_policy.permits(context, principals,
                                                 permission)
            @asyncio.coroutine
            def _secured_view(context, request):
                result = yield from _permitted(context, request)
                if result:
                    return view(context, request)
                view_name = getattr(view, '__name__', view)
                msg = getattr(
                    request, 'authdebug_message',
                    'Unauthorized: %s failed permission check' % view_name)
                raise HTTPForbidden(msg, result=result)
            _secured_view.__call_permissive__ = view
            _secured_view.__permitted__ = _permitted
            _secured_view.__permission__ = permission
            wrapped_view = _secured_view

        return wrapped_view

    def _rendered_view(self, view, view_renderer):
        """ Render an async view """

        @asyncio.coroutine
        def rendered_view(context, request):

            renderer = view_renderer
            result = view(context, request)
            if is_generator(result):
                result = yield from result

            if result.__class__ is Response:  # potential common case
                response = result
            else:
                registry = self.registry
                # this must adapt, it can't do a simple interface check
                # (avoid trying to render webob responses)
                response = registry.queryAdapterOrSelf(result, IResponse)
                if response is None:
                    attrs = getattr(request, '__dict__', {})
                    if 'override_renderer' in attrs:
                        # renderer overridden by newrequest event or other
                        renderer_name = attrs.pop('override_renderer')
                        renderer = renderers.RendererHelper(
                            name=renderer_name,
                            package=self.kw.get('package'),
                            registry=registry)
                    if '__view__' in attrs:
                        view_inst = attrs.pop('__view__')
                    else:
                        view_inst = getattr(view, '__original_view__', view)
                    response = renderer.render_view(request, result, view_inst,
                                                    context)
            return response

        return rendered_view

    def _response_resolved_view(self, view):
        registry = self.registry

        def viewresult_to_response(context, request):

            result = yield from view(context, request)
            if result.__class__ is Response:  # common case
                response = result
            else:
                response = registry.queryAdapterOrSelf(result, IResponse)
                if response is None:
                    if result is None:
                        append = (' You may have forgotten to return a value '
                                  'from the view callable.')
                    elif isinstance(result, dict):
                        append = (' You may have forgotten to define a '
                                  'renderer in the view configuration.')
                    else:
                        append = ''

                    msg = ('Could not convert return value of the view '
                           'callable %s into a response object. '
                           'The value returned was %r.' + append)

                    raise ValueError(msg % (view_description(view), result))

            return response

        return viewresult_to_response


def includeme(config):
    # Does not work as expected :/
    # config.registry.registerUtility(Tweens(), ITweens)

    config.add_directive('add_coroutine_view', add_coroutine_view)
    config.add_directive('make_asyncio_app', make_asyncio_app)
    config.add_directive('add_exit_handler', add_exit_handler)



import asyncio

try:
    from asyncio_redis import Connection
    from pyramid_kvs import serializer, kvs, session, cache
    from pyramid_kvs.serializer import serializer
except ImportError:
    raise Exception("Packages asyncio_redis and pyramid_kvs required")


class AIORedis(kvs.KVS):

    @asyncio.coroutine
    def get(self, key, default=None):
        if key is None:
            return default
        ret = yield from self.raw_get(key)
        if ret is None:
            return default
        return self._serializer.loads(ret)

    @asyncio.coroutine
    def set(self, key, value, ttl=None):
        value = self._serializer.dumps(value)
        return (yield from self.raw_set(key, value, ttl or self.ttl))

    @asyncio.coroutine
    def delete(self, key):
        yield from self._client.delete([self._get_key(key)])

    @asyncio.coroutine
    def _create_client(self, **kwargs):
        return (yield from Connection.create(**kwargs))

    def _get_key(self, key):
        return super()._get_key(key).decode('utf-8')

    @asyncio.coroutine
    def raw_get(self, key, default=None):
        ret = yield from self._client.get(self._get_key(key))
        return default if ret is None else ret

    @asyncio.coroutine
    def raw_set(self, key, value, ttl):
        yield from self._client.setex(self._get_key(key), ttl,
                                      value)

    @asyncio.coroutine
    def incr(self, key):
        return (yield from self._client.incr(self._get_key(key)))


kvs._implementations['aioredis'] = AIORedis  # XXX Private access




import asyncio

import simplejson as json
from pyramid.events import NewRequest

from .kvs import cache, serializer


class ApplicationCache(cache.ApplicationCache):

    @classmethod
    @asyncio.coroutine
    def connect(cls, settings):
        """ Call that method in the pyramid configuration phase.
        """
        
        kvs_cache = json.loads(settings['kvs.cache'])
        kvs_cache['kvs'] = 'aioredis'
        settings['kvs.cache'] = json.dumps(kvs_cache)
        super().connect(settings)
        cls.client._client = yield from cls.client._client

    @asyncio.coroutine
    def get(self, key, default=None):
        return (yield from self.client.get(key, default))

    @asyncio.coroutine
    def set(self, key, value, ttl=None):
        yield from self.client.set(key, value, ttl=None)

    @asyncio.coroutine
    def pop(self, key, default=None):
        try:
            ret = yield from self.get(key)
            yield from self.client.delete(key)
            return ret
        except KeyError:
            return default


def subscribe_cache(event):
    request = event.request
    request.set_property(ApplicationCache(request), 'cache', reify=True)


@asyncio.coroutine
def includeme(config):
    settings = config.registry.settings
    if 'kvs.cache' in settings:
        yield from ApplicationCache.connect(settings)
        config.add_subscriber(subscribe_cache, NewRequest)



import os
import sys
import re

from setuptools import setup, find_packages

NAME = 'pyramid-asyncio'

py_version = sys.version_info[:2]
if py_version < (3, 3):
    raise Exception("{name} requires Python >= 3.3.".format(name=NAME))

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, 'README.rst')) as readme:
    README = readme.read()
with open(os.path.join(here, 'CHANGES.rst')) as changes:
    CHANGES = changes.read()

with open(os.path.join(here, NAME.replace('-', '_'),
                       '__init__.py')) as version:
    VERSION = re.compile(r".*__version__ = '(.*?)'",
                         re.S).match(version.read()).group(1)


requires = [
    'pyramid',
    'gunicorn >= 19.0',
    'aiohttp',
    ]

extras_require = {
    'session': ['pyramid-kvs >= 0.2',  # XXX unreleased
                'asyncio-redis',
                'simplejson'
                ]
}

if py_version < (3, 4):
    requires.append('asyncio')


setup(name=NAME,
      version=VERSION,
      description='Pyramid Asyncio Glue',
      # long_description=README + '\n\n' + CHANGES,
      classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Framework :: Pyramid",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: WSGI :: Application",
        "Intended Audience :: Developers",
        "License :: Repoze Public License",
        ],
      author='Guillaume Gauvrit',
      author_email='guillaume@gauvr.it',
      url='https://github.com/mardiros/pyramid_asyncio',
      keywords='pyramid asyncio',
      packages=find_packages(),
      include_package_data=True,
      zip_safe=False,
      test_suite='{name}.tests'.format(name=NAME),
      install_requires=requires,
      license="BSD-derived (http://www.repoze.org/LICENSE.txt)",
      extras_require=extras_require,
      entry_points = """\
      [pyramid.scaffold]
      aio_jinja2=pyramid_asyncio.scaffolds:AioJinja2Template
      """
      )



import asyncio

from gunicorn.workers.gaiohttp import AiohttpWorker as BaseWorker

class AiohttpWorker(BaseWorker):

    @asyncio.coroutine
    def _run(self):

        if hasattr(self.wsgi, 'open'):
            yield from self.wsgi.open()

        yield from super()._run()

    def handle_quit(self, sig, frame):
        self.alive = False

    handle_exit = handle_quit
